@@ -120,6 +120,9 @@
 (defun oddp (num)
   (= 1 (mod num 2)))
 
+(defun rationalize (num)
+  (rational num))
+
 (defun gcd (&rest nums)
   (let ((l (length nums)))
     (cond
@@ -547,6 +550,15 @@
 
 
 
+(defun (setf subseq) (newval seq start &optional end)
+  (let ((len (- (or end (length seq)) start))
+	(newvallen (length newval)))
+    (dotimes (i (min len newvallen))
+      (setf (elt seq (+ i start)) (elt newval i)))
+    newval))
+
+
+
 (define-setf-expander symbol-value (sym)
   (let ((var (gensym))
 	(val (gensym)))
@@ -558,6 +570,52 @@
      `(symbol-value ,var))))
 
 
+(define-setf-expander values (&rest places)
+  (let* ((exps (mapcar (lambda (p) (multiple-value-list (get-setf-expansion p))) places))
+	 (varsyms (apply 'append (mapcar #'car exps)))
+	 (valforms (apply 'append (mapcar #'cadr exps)))
+	 (storesyms (mapcar #'caaddr exps))
+	 (otherstoresyms (apply 'append (mapcar #'cdaddr exps))))
+    (values
+     varsyms
+     valforms
+     storesyms
+     `(let ,otherstoresyms ,@(mapcar (lambda (e) (nth 3 e)) exps) (values ,@storesyms))
+     `(values ,@(mapcar (lambda (e) (nth 4 e)) exps)))))
+
+
+
+(defmacro shiftf (&rest args)
+  (let* ((places (subseq args 0 (- (length args) 1)))
+	 (exps (mapcar (lambda (p) (multiple-value-list (get-setf-expansion p))) places))
+	 (varsyms (apply 'append (mapcar #'car exps)))
+	 (valforms (apply 'append (mapcar #'cadr exps)))
+	 (storesyms (mapcar #'caaddr exps))
+	 (otherstoresyms (apply 'append (mapcar #'cdaddr exps)))
+	 (oldval (gensym))
+	 (newval (gensym)))
+    `(let* (,@(mapcar #'list varsyms valforms)
+	    (,newval ,(car (last args)))
+	      ,@(mapcar #'list storesyms (append (mapcar (lambda (e) (nth 4 e)) (cdr exps)) `(,newval)))
+	      ,@otherstoresyms
+	      (,oldval ,(nth 4 (car exps))))
+       ,@(mapcar #'cadddr exps)
+       ,oldval)))
+
+
+(defmacro rotatef (&rest places)
+  (let* ((exps (mapcar (lambda (p) (multiple-value-list (get-setf-expansion p))) places))
+	 (varsyms (apply 'append (mapcar #'car exps)))
+	 (valforms (apply 'append (mapcar #'cadr exps)))
+	 (storesyms (mapcar #'caaddr exps))
+	 (otherstoresyms (apply 'append (mapcar #'cdaddr exps))))
+    `(let* (,@(mapcar #'list varsyms valforms)
+	    ,@(mapcar #'list storesyms (append (mapcar (lambda (e) (nth 4 e)) (cdr exps)) `(,(nth 4 (car exps)))))
+	      ,@otherstoresyms)
+       ,@(mapcar #'cadddr exps)
+       nil)))
+
+
 
 (defmacro defsetf (accessfn secondarg &rest args)
   (if (listp secondarg)
@@ -602,14 +660,14 @@
 	    (setq args (cons l args)))))
     `(defmacro ,name (place . ,lambda-list)
        (let ((exp (multiple-value-list (get-setf-expansion place)))
-	      letf)
-	  (dolist (f (car exp))
-	    (setq letf (cons (list f (caadr exp)) letf))
-	    (setf (cadr exp) (cdadr exp)))
-	  (setq letf (cons (list (caaddr exp) (list ',func (nth 4 exp) ,@args)) letf))
-	  (setq letf (list 'let* (reverse letf) nil))
-	  (setf (caddr letf) (nth 3 exp))
-	  letf))))
+	     letf)
+	 (dolist (f (car exp))
+	   (setq letf (cons (list f (caadr exp)) letf))
+	   (setf (cadr exp) (cdadr exp)))
+	 (setq letf (cons (list (caaddr exp) (list ',func (nth 4 exp) ,@args)) letf))
+	 (setq letf (list 'let* (reverse letf) nil))
+	 (setf (caddr letf) (nth 3 exp))
+	 letf))))
 
 
 (define-modify-macro incf (&optional (delta 1)) +)
@@ -617,38 +675,33 @@
 (define-modify-macro decf (&optional (delta 1)) -)
 
 
-(defmacro and (&rest forms)
-  (if (not forms)
-      t
-      (let ((valsym (gensym)))
-	`(let ((,valsym (multiple-value-list ,(car forms))))
-           (if (car ,valsym)
-               ,(if (not (cdr forms))
-                    `(values-list ,valsym)
-                    `(and ,@(cdr forms))))))))
-
-
-(defmacro or (&rest forms)
-  (if (not forms)
-      nil
-      (let ((valsym (gensym)))
-	`(let ((,valsym (multiple-value-list ,(car forms))))
-	   (if (car ,valsym)
-	       ,(if (not (cdr forms))
-		    `(values-list ,valsym)
-		    `(car ,valsym))
-               (or ,@(cdr forms)))))))
-
 
 (defmacro cond (&body body)
-  (when body
-    (let ((first (car body))
-	  (rest (cdr body))
-	  (valsym (gensym)))
-      `(let ((,valsym ,(car first)))
-	 (if ,valsym
-	     (progn ,valsym ,@(cdr first))
-	     (cond ,@rest))))))
+  (let (ret l)
+    (dolist (f body)
+      (if (cdr f)
+	  (progn
+	    (let ((cl
+		   `(if ,(car f)
+			,(if (cddr f)
+			     `(progn ,@(cdr f))
+			     (if (cdr f)
+				 (cadr f)))
+			,nil)))
+	      (if ret
+		  (setf (car l) cl l (cdddr cl))
+		  (setf ret cl l (cdddr cl)))))
+	  (progn
+	    (let* ((s (gensym))
+		   (cl
+		    `(let ((,s ,(car f)))
+		       (if ,s
+			   ,s
+			   ,nil))))
+	      (if ret
+		  (setf (car l) cl l (cddr (cdaddr cl)))
+		  (setf ret cl l (cddr (cdaddr cl))))))))
+    ret))
 
 
 
@@ -739,6 +792,7 @@
   `(return-from nil (values-list (multiple-value-list ,val))))
 
 
+
 (defmacro multiple-value-bind (vars valform &body forms)
   (let ((restsym (gensym)))
     `(multiple-value-call (lambda (&optional ,@vars &rest ,restsym)
@@ -746,6 +800,18 @@
        ,valform)))
 
 
+(defmacro multiple-value-setq (vars valform)
+  (let ((valssym (gensym)))
+    `(let ((,valssym (multiple-value-list ,valform)))
+       ,@(let (setqs (i 0))
+	   (dolist (v vars)
+	     (setq setqs (cons `(setq ,v (nth ,i ,valssym)) setqs))
+	     (incf i))
+	   (reverse setqs))
+       (car ,valssym))))
+
+
+
 (defmacro prog (bins &body body)
   `(let ,bins
      (block nil
@@ -760,6 +826,13 @@
 	  ,@body))))
 
 
+(defmacro multiple-value-prog1 (form &body otherforms)
+  (let ((formvalsym (gensym)))
+    `(let ((,formvalsym (multiple-value-list ,form)))
+       ,@otherforms
+       (values-list ,formvalsym))))
+
+
 
 (defun every (pred &rest sequences)
   (let ((n (apply 'min (mapcar #'length sequences))))
@@ -1389,6 +1462,30 @@
 	    (setq propl (cddr propl)))))))
 
 
+(defun getf (plist ind &optional def)
+  (let ((l (length plist)))
+    (do ((i 0 (+ i 2)))
+	((>= i l))
+      (if (eq (car plist) ind)
+	  (return-from getf (cadr plist)))
+      (setq plist (cddr plist))))
+  def)
+
+
+(defun (setf getf) (newval plist ind &optional def)
+  (let ((l (length plist)))
+    (do ((i 0 (+ i 2)))
+	((>= i l))
+      (when (eq (car plist) ind)
+	(setf (cadr plist) newval)
+	(return-from getf newval))
+      (unless (cddr plist)
+	(setf (cddr plist) (list ind newval))
+	(return-from getf newval))
+      (setq plist (cddr plist))))
+  def)
+
+
 
 (defun char (str ind)
   (aref str ind))
@@ -1839,7 +1936,7 @@
 
 
 
-(defmacro assert (test &optional places (conddat "Assertion does not hold") condargs)
+(defmacro assert (test &optional places (conddat "Assertion does not hold") &rest condargs)
   `(unless ,test
      (apply 'error ,conddat ,condargs)))
 
@@ -1967,9 +2064,10 @@
 		      (+ start i))))
 	   (if (= i 0)
 	       (if initvalprov
-		   (setq out (funcall fun initial-value (funcall key (elt seq j))))
-		   (progn
-		     (setq out (funcall key (elt seq j)))))
+		   (setq out (if from-end
+				 (funcall fun (funcall key (elt seq j)) initial-value)
+				 (funcall fun initial-value (funcall key (elt seq j)))))
+		   (setq out (funcall key (elt seq j))))
 	       (if from-end
 		   (setq out (funcall fun (funcall key (elt seq j)) out))
 		   (setq out (funcall fun out (funcall key (elt seq j))))))))
@@ -2017,14 +2115,14 @@
   (pathname fn))
 
 
+(defun translate-logical-pathname (pn &key)
+  (pathname pn))
+
+
 (defun pathname-host (fn &key case))
 
 (defun pathname-device (fn &key case))
 
-(defun pathname-directory (fn &key case))
-
-(defun pathname-type (fn &key case))
-
 (defun pathname-version (fn &key case))
 
 
@@ -2040,7 +2138,16 @@
 
 
 (defun merge-pathnames (pn &optional defpn defv)
-  (pathname pn))
+  (let* ((p1 (pathname pn))
+	 (p2 (pathname defpn))
+	 (dir (or (pathname-directory p1) (pathname-directory p2)))
+	 (dirtype (if (and
+		       (> (length dir) 0) (char= (elt dir 0) #\/))
+		      :absolute
+		      :relative))
+	 (name (or (pathname-name p1) (pathname-name p2))))
+    (make-pathname :directory (if dir (list dirtype dir))
+		   :name name)))
 
 
 
@@ -2048,7 +2155,9 @@
 
 (defun file-write-date (fn))
 
-(defun user-homedir-pathname (&optional h))
+(defun user-homedir-pathname (&optional h)
+  (make-pathname :directory (cl-user:al-getenv "HOME")))
+
 
 
 
@@ -2116,6 +2225,82 @@
     (get-output-stream-string out)))
 
 
+
+(defun force-output (&optional (str *standard-output*))
+  (finish-output str))
+
+
+
+(defmacro with-input-from-string ((var strform &key index (start 0) end) &body forms)
+  (let ((strsym (gensym)))
+    `(let* ((,strsym ,strform)
+	    (,var (make-string-input-stream ,strsym ,start ,end)))
+       (unwind-protect
+	    (progn
+	      ,@(if index
+		    `((setq ,index (length ,strsym))))
+	      ,@forms)
+	 ,@(if index
+	       `((setq ,index (- ,index (length (cl-user:al-string-input-stream-string ,var))))))
+	 (close ,var)))))
+
+
+(defmacro with-output-to-string ((var &optional strform &key element-type) &body forms)
+  (let ((blockname (gensym)))
+    `(block ,blockname
+       (let ((,var (make-string-output-stream ,strform)))
+	 (unwind-protect
+	      (progn
+		,@forms)
+	   ,(if strform
+		`(close ,var)
+		`(let ((out (get-output-stream-string ,var)))
+		   (close ,var)
+		   (return-from ,blockname out))))))))
+
+
+
+(defun pprint (obj &optional (out *standard-output*))
+  (terpri out)
+  (let ((*print-pretty* t))
+    (prin1 obj out))
+  (values))
+
+
+(defun indent (str n)
+  (dotimes (i n)
+    (write-string "  " str)))
+
+(defun print-object-nicely (str obj)
+  (indent str cl-user::*al-pprint-depth*)
+  (let (*print-pretty*)
+    (if (atom obj)
+	(format str "~a" obj)
+	(format str "(~a" (car obj))))
+  (when (consp obj)
+    (incf cl-user::*al-pprint-depth*)
+    (setq obj (cdr obj))
+    (unless obj
+      (let (*print-pretty*)
+	(format str ")")))
+    (dotimes (i (length obj))
+      (if (atom (car obj))
+	  (let (*print-pretty*)
+	    (format str " ~a" (car obj)))
+	  (format str "~%~a" (car obj)))
+      (if (not (cdr obj))
+	  (let (*print-pretty*)
+	    (format str ")"))
+	  (if (atom (cdr obj))
+	      (let (*print-pretty*)
+		(format str " . ~a)" (cdr obj)))))
+      (setq obj (cdr obj)))
+    (decf cl-user::*al-pprint-depth*)))
+
+(setq *print-pprint-dispatch* (cons (list 't #'print-object-nicely 0) *print-pprint-dispatch*))
+
+
+
 (defmacro do-all-symbols (varres &body body)
   (let ((packsym (gensym)))
     `(progn
@@ -2126,27 +2311,71 @@
 	 ,(cadr varres)))))
 
 
+(defun find-all-symbols (sym)
+  (let ((packs (list-all-packages))
+	(str (string sym))
+	out)
+    (dolist (p packs (remove-duplicates out :test #'eq))
+      (let ((s (find-symbol str p)))
+	(if s
+	    (setq out (cons s out)))))))
 
-(defun loop-parse-accumulation (forms)
+
+
+(defmacro with-slots (slots inst &rest forms)
+  (let ((instsym (gensym)))
+    `(let ((,instsym ,inst))
+       (symbol-macrolet
+	   ,(let (out)
+	      (dolist (s slots)
+		(if (consp s)
+		    (setq out (cons `(,(car s) (slot-value ,instsym ',(cadr s))) out))
+		    (setq out (cons `(,s (slot-value ,instsym ',s)) out))))
+	      out)
+	 ,@forms))))
+
+
+(defmacro with-accessors (slots inst &rest forms)
+  (let ((instsym (gensym)))
+    `(let ((,instsym ,inst))
+       (symbol-macrolet
+	   ,(let (out)
+	      (dolist (s slots)
+		(setq out (cons `(,(car s) (,(cadr s) ,instsym)) out)))
+	      out)
+	 ,@forms))))
+
+
+
+(defun loop-parse-accumulation (forms ifclvar)
   (let ((sym (string (car forms)))
 	var
-	out)
+	(val (if (and (symbolp (cadr forms))
+		      (string= (cadr forms) "IT"))
+		 ifclvar
+		 (cadr forms)))
+	out
+	init)
     (if (string= (string (caddr forms)) "INTO")
 	(setq var (cadddr forms))
 	(setq var (gensym)))
     (cond
       ((or (string= sym "COLLECT")
 	   (string= sym "COLLECTING"))
-       (setq out `(setq ,var (append ,var `(,,(cadr forms))))))
+       (setq out `(setq ,var (append ,var (list ,val)))))
       ((or (string= sym "APPEND")
 	   (string= sym "APPENDING"))
-       (setq out `(setq ,var (append ,var ,(cadr forms)))))
+       (setq out `(setq ,var (append ,var ,val))))
       ((or (string= sym "NCONC")
 	   (string= sym "NCONCING"))
-       (setq out `(setq ,var (nconc ,var ,(cadr forms))))))
+       (setq out `(setq ,var (nconc ,var ,val))))
+      ((or (string= sym "SUM")
+	   (string= sym "SUMMING"))
+       (setq init 0)
+       (setq out `(setq ,var (+ ,var ,val)))))
     (if (string= (string (caddr forms)) "INTO")
-	(values out var nil (cddddr forms))
-	(values out nil var (cddr forms)))))
+	(values out var nil init (cddddr forms))
+	(values out nil var init (cddr forms)))))
 
 
 (defun loop-parse-iteration (forms)
@@ -2160,51 +2389,51 @@
 	    (cond
 	      ((string= sym "FROM")
 	       (if (= (length iter) 1)
-		   (setq iter `(:from ,(car iter) nil 0 nil 1 1)))
+		   (setq iter `(:from ,(car iter) nil 0 nil 1 ,1)))
 	       (setf (elt iter 3) (cadr f))
 	       (setq f (cddr f)))
 	      ((string= sym "DOWNFROM")
 	       (if (= (length iter) 1)
-		   (setq iter `(:from ,(car iter) nil 0 nil 1 1)))
+		   (setq iter `(:from ,(car iter) nil 0 nil 1 ,1)))
 	       (setf (elt iter 3) (cadr f))
 	       (setf (elt iter 5) -1)
 	       (setq f (cddr f)))
 	      ((string= sym "UPFROM")
 	       (if (= (length iter) 1)
-		   (setq iter `(:from ,(car iter) nil 0 nil 1 1)))
+		   (setq iter `(:from ,(car iter) nil 0 nil 1 ,1)))
 	       (setf (elt iter 3) (cadr f))
 	       (setf (elt iter 5) 1)
 	       (setq f (cddr f)))
 	      ((string= sym "TO")
 	       (if (= (length iter) 1)
-		   (setq iter `(:from ,(car iter) nil 0 nil 1 1)))
+		   (setq iter `(:from ,(car iter) nil 0 nil 1 ,1)))
 	       (setf (elt iter 4) (cadr f))
 	       (setq f (cddr f)))
 	      ((string= sym "DOWNTO")
 	       (if (= (length iter) 1)
-		   (setq iter `(:from ,(car iter) nil 0 nil 1 1)))
+		   (setq iter `(:from ,(car iter) nil 0 nil 1 ,1)))
 	       (setf (elt iter 4) (cadr f))
 	       (setf (elt iter 5) -1)
 	       (setq f (cddr f)))
 	      ((string= sym "UPTO")
                (if (= (length iter) 1)
-		   (setq iter `(:from ,(car iter) nil 0 nil 1 1)))
+		   (setq iter `(:from ,(car iter) nil 0 nil 1 ,1)))
 	       (setf (elt iter 4) (cadr f))
 	       (setf (elt iter 5) 1)
 	       (setq f (cddr f)))
 	      ((string= sym "BELOW")
 	       (if (= (length iter) 1)
-		   (setq iter `(:from ,(car iter) nil 0 nil 1 1)))
+		   (setq iter `(:from ,(car iter) nil 0 nil 1 ,1)))
 	       (setf (elt iter 4) (1- (cadr f)))
 	       (setq f (cddr f)))
 	      ((string= sym "ABOVE")
 	       (if (= (length iter) 1)
-		   (setq iter `(:from ,(car iter) nil 0 nil 1 1)))
+		   (setq iter `(:from ,(car iter) nil 0 nil 1 ,1)))
 	       (setf (elt iter 4) (1+ (cadr f)))
 	       (setq f (cddr f)))
 	      ((string= sym "BY")
 	       (if (= (length iter) 1)
-		   (setq iter `(:from ,(car iter) nil 0 nil 1 1)))
+		   (setq iter `(:from ,(car iter) nil 0 nil 1 ,1)))
 	       (cond
 		 ((eq (car iter) :from)
 		  (setf (elt iter 6) (cadr f)))
@@ -2212,13 +2441,13 @@
 		  (setf (elt iter 4) (cadr f))))
 	       (setq f (cddr f)))
 	      ((string= sym "IN")
-	       (setq iter `(:in ,(car iter) nil ,(cadr f) #'cdr))
+	       (setq iter `(:in ,(car iter) nil ,(cadr f) ,#'cdr))
 	       (setq f (cddr f)))
 	      ((string= sym "ON")
-	       (setq iter `(:on ,(car iter) nil ,(cadr f) #'cdr))
+	       (setq iter `(:on ,(car iter) nil ,(cadr f) ,#'cdr))
 	       (setq f (cddr f)))
 	      ((string= sym "=")
-	       (setq iter `(:eq ,(car iter) ,(cadr f) nil))
+	       (setq iter `(:eq ,(car iter) ,(cadr f) ,nil))
 	       (setq f (cddr f)))
 	      ((string= sym "THEN")
 	       (setf (elt iter 3) (cadr f))
@@ -2227,7 +2456,7 @@
 	       (setq iter `(:across ,(car iter) nil nil ,(cadr f)))
 	       (setq f (cddr f)))
 	      ((string= sym "BEING")
-	       (setq iter `(:hashtable ,(car iter) nil nil nil))
+	       (setq iter `(:hashtable ,(car iter) nil nil ,nil))
 	       (setq f (cdr f))
 	       (if (or (string= (string (car f)) "EACH")
 		       (string= (string (car f)) "THE"))
@@ -2276,14 +2505,15 @@
 	retst
 	vars
 	returnvar
+	(ifclvar (gensym))
 	(sym (string (car forms))))
     (cond
       ((or (string= sym "IF")
 	   (string= sym "WHEN"))
-       (setq out `(if ,(cadr forms) nil))
+       (setq out `(let ((,ifclvar ,(cadr forms))) (if ,ifclvar ,nil)))
        (setq forms (cddr forms)))
       ((string= sym "UNLESS")
-       (setq out `(if (not ,(cadr forms)) nil))
+       (setq out `(let ((,ifclvar ,(cadr forms))) (if (not ,ifclvar) ,nil)))
        (setq forms (cddr forms))))
     (do ((f forms))
 	((not f) (setq forms f))
@@ -2293,25 +2523,29 @@
 	      ((or (string= sym "IF")
 		   (string= sym "WHEN")
 		   (string= sym "UNLESS"))
-	       (multiple-value-bind (res frm vs rv)
+	       (multiple-value-bind (res frm vs rv ifcv)
 		   (loop-parse-conditional f loopname)
-		 (if (= (length out) 3)
+		 (if (= (length (caddr out)) 3)
 		     (setq docl (cons res docl))
 		     (setq elsecl (cons res elsecl)))
 		 (setq vars (append vs vars))
 		 (if rv
 		     (setq returnvar rv))
+		 (setq ifclvar ifcv)
 		 (setq f frm)))
 	      ((string= sym "ELSE")
-	       (setq out (append out (list nil)))
+	       (setf (caddr out) (append (caddr out) (list nil)))
 	       (setq f (cdr f)))
 	      ((or (string= sym "DO")
 		   (string= sym "DOING")
 		   (string= sym "AND"))
 	       (setq f (cdr f)))
 	      ((string= sym "RETURN")
-	       (setq retst `(return-from ,loopname ,(cadr f)))
-	       (if (= (length out) 3)
+	       (setq retst `(return-from ,loopname ,(if (and (symbolp (cadr f))
+							     (string= (cadr f) "IT"))
+							ifclvar
+							(cadr f))))
+	       (if (= (length (caddr out)) 3)
 		   (setq docl (cons retst docl))
 		   (setq elsecl (cons retst elsecl)))
 	       (setq f (cddr f)))
@@ -2320,12 +2554,14 @@
 		   (string= sym "APPEND")
 		   (string= sym "APPENDING")
 		   (string= sym "NCONC")
-		   (string= sym "NCONCING"))
-	       (multiple-value-bind (res var1 var2 frm) (loop-parse-accumulation f)
-		 (if (= (length out) 3)
+		   (string= sym "NCONCING")
+		   (string= sym "SUM")
+		   (string= sym "SUMMING"))
+	       (multiple-value-bind (res var1 var2 init frm) (loop-parse-accumulation f ifclvar)
+		 (if (= (length (caddr out)) 3)
 		     (setq docl (cons res docl))
 		     (setq elsecl (cons res elsecl)))
-		 (setq vars (cons (list (or var1 var2)) vars))
+		 (setq vars (cons (list (list (or var1 var2) init)) vars))
 		 (if var2
 		     (setq returnvar var2))
 		 (setq f frm)))
@@ -2336,18 +2572,18 @@
 	       (setq forms f)
 	       (return nil))))
 	  (progn
-	    (if (= (length out) 3)
+	    (if (= (length (caddr out)) 3)
 		(setq docl (cons (car f) docl))
 		(setq elsecl (cons (car f) elsecl)))
 	    (setq f (cdr f)))))
     (setq docl (reverse docl))
     (setq docl (cons 'progn docl))
-    (setf (elt out 2) docl)
-    (when (= (length out) 4)
+    (setf (elt (caddr out) 2) docl)
+    (when (= (length (caddr out)) 4)
       (setq elsecl (reverse elsecl))
       (setq elsecl (cons 'progn elsecl))
-      (setf (elt out 3) elsecl))
-    (values out forms vars returnvar)))
+      (setf (elt (caddr out) 3) elsecl))
+    (values out forms vars returnvar ifclvar)))
 
 
 (defun flatten-tree-skipping-nils (tree)
@@ -2382,7 +2618,9 @@
 	lets l
 	initial-setup iteration-setup
 	out o
-	returnvar)
+	returnvar
+	ifclvar
+	defaultret)
     (do ()
 	((not forms))
       (let ((form (car forms)))
@@ -2401,7 +2639,7 @@
 		 (when parvars
 		   (setq vars (cons parvars vars))
 		   (setq parvars nil))
-		 (setq var `(,(cadr forms) nil))
+		 (setq var `(,(cadr forms) ,nil))
 		 (if (and (symbolp (caddr forms))
 			  (string= (caddr forms) "="))
 		     (progn
@@ -2410,7 +2648,7 @@
 		     (setq forms (cddr forms)))
 		 (setq parvars (cons var parvars)))
 		((string= sym "AND")
-		 (setq var `(,(cadr forms) nil))
+		 (setq var `(,(cadr forms) ,nil))
 		 (if (and (symbolp (caddr forms))
 			  (string= (caddr forms) "="))
 		     (progn
@@ -2421,12 +2659,13 @@
 		((or (string= sym "IF")
 		     (string= sym "WHEN")
 		     (string= sym "UNLESS"))
-		 (multiple-value-bind (res frm vs rv)
+		 (multiple-value-bind (res frm vs rv ifcv)
 		     (loop-parse-conditional forms block-name)
 		   (setq do-forms (append do-forms `(,res)))
 		   (setq vars (append vs vars))
 		   (if rv
 		       (setq returnvar rv))
+		   (setq ifclvar ifcv)
 		   (setq forms frm)))
 		((or (string= sym "DO")
 		     (string= sym "DOING"))
@@ -2442,15 +2681,34 @@
 		 (multiple-value-bind (res frm) (loop-parse-iteration forms)
 		   (setq iters (append iters `(,res)))
 		   (setq forms frm)))
+		((string= sym "REPEAT")
+		 (setq iters (append iters `((:repeat ,(gensym) ,(gensym) ,(cadr forms)))))
+		 (setq forms (cddr forms)))
+		((string= sym "ALWAYS")
+		 (setq do-forms (append do-forms `((if (not ,(cadr forms)) (return-from ,block-name ,nil)))))
+		 (setq defaultret t)
+		 (setq forms (cddr forms)))
+		((string= sym "NEVER")
+		 (setq do-forms (append do-forms `((if ,(cadr forms) (return-from ,block-name ,nil)))))
+		 (setq defaultret t)
+		 (setq forms (cddr forms)))
+		((string= sym "WHILE")
+		 (setq do-forms (append do-forms `((if (not ,(cadr forms)) (return-from ,inner-block-name ,nil)))))
+		 (setq forms (cddr forms)))
+		((string= sym "UNTIL")
+		 (setq do-forms (append do-forms `((if ,(cadr forms) (return-from ,inner-block-name ,nil)))))
+		 (setq forms (cddr forms)))
 		((or (string= sym "COLLECT")
 		     (string= sym "COLLECTING")
 		     (string= sym "APPEND")
 		     (string= sym "APPENDING")
 		     (string= sym "NCONC")
-		     (string= sym "NCONCING"))
-		 (multiple-value-bind (res var1 var2 frm) (loop-parse-accumulation forms)
+		     (string= sym "NCONCING")
+		     (string= sym "SUM")
+		     (string= sym "SUMMING"))
+		 (multiple-value-bind (res var1 var2 init frm) (loop-parse-accumulation forms ifclvar)
 		   (setq do-forms (append do-forms `(,res)))
-		   (setq vars (cons (list (or var1 var2)) vars))
+		   (setq vars (cons (list (list (or var1 var2) init)) vars))
 		   (if var2
 		       (setq returnvar var2))
 		   (setq forms frm)))
@@ -2464,12 +2722,12 @@
 	(setq vars (cons parvars vars)))
     (setq vars (reverse vars))
 
-    (setq lets `(let (,skip-body-sym) nil))
+    (setq lets `(let (,skip-body-sym) ,nil))
     (setq l (cddr lets))
 
     (dolist (v vars)
-      (setf (car l) `(let ,v nil))
-      (setq l (cddar l)))
+      (setf (car l) `(cl-user:al-loopy-destructuring-bind ,(mapcar 'car v) ,(cons 'list (mapcar 'cadr v)) ,nil))
+      (setq l (cdddar l)))
 
     (let (vrs)
       (dolist (i iters)
@@ -2531,6 +2789,16 @@
 				      (return-from ,inner-block-name nil))
 				  (incf ,(elt i 2))
 				  . ,iteration-setup)))
+	(when (eq (car i) :repeat)
+	  (setq vrs (list* (elt i 1) (elt i 2) vrs))
+	  (setq initial-setup `((if (>= ,(elt i 1) ,(elt i 2))
+				    (setq ,skip-body-sym t))
+				(setq ,(elt i 1) 0 ,(elt i 2) ,(elt i 3))
+				. ,initial-setup))
+	  (setq iteration-setup `((if (>= ,(elt i 1) ,(elt i 2))
+				      (return-from ,inner-block-name nil))
+				  (incf ,(elt i 1))
+				  . ,iteration-setup)))
 	(when (eq (car i) :hashtable)
 	  (setq initial-setup `((cl-user:al-loopy-setq ,(elt i 2) (cdar ,(elt i 3)))
 				(cl-user:al-loopy-setq ,(elt i 1) (caar ,(elt i 3)))
@@ -2547,28 +2815,32 @@
 				  . ,iteration-setup))))
 
       (when vrs
-	(setf (car l) `(let ,vrs nil))
+	(setf (car l) `(let ,vrs ,nil))
 	(setq l (cddar l)))
 
       (let* ((out `(block ,block-name
-		     nil))
+		     ,nil))
 	     (o (cddr out)))
 	(setf (car o) lets)
 	(setq o l)
-	(setf (car o) `(tagbody
-			  (progn ,@(reverse initial-setup))
-			  ,prologue-tag
-			  ,@initially-forms
-			  (unless ,skip-body-sym
-			    (block ,inner-block-name
-			      (tagbody
-				 ,body-tag
-				 ,@do-forms
-				 (progn ,@(reverse iteration-setup))
-				 (go ,body-tag))))
-			  ,@finally-forms
-			  ,(if returnvar
-			       `(return-from ,block-name ,returnvar))))
+	(setf (car o) `(progn
+			 (tagbody
+			    ,@(if initial-setup
+				  `((progn ,@(reverse initial-setup))))
+			    ,prologue-tag
+			    ,@initially-forms
+			    (unless ,skip-body-sym
+			      (block ,inner-block-name
+				(tagbody
+				   ,body-tag
+				   ,@do-forms
+				   ,@(if iteration-setup
+					 `((progn ,@(reverse iteration-setup))))
+				   (go ,body-tag))))
+			    ,@finally-forms)
+			 ,(if returnvar
+			      `(return-from ,block-name ,returnvar)
+			      defaultret)))
 	out))))
 
 
@@ -2648,6 +2920,16 @@
 
 
 
+(defmacro with-compilation-unit (opts &rest forms)
+  `(progn
+     ,@forms))
+
+
+
+(defparameter *readtable* nil)
+
+
+
 (defmacro with-standard-io-syntax (&body forms)
   `(let ((*package* (find-package 'cl-user))
 	 (*print-array* t)
@@ -2725,13 +3007,33 @@
 
 
 
-(defun break (&optional form &rest args)
+(defun cerror (format datum &rest args)
   (with-simple-restart (continue "")
-    (invoke-debugger (make-condition 'simple-condition)))
+    (apply 'error datum args))
   nil)
 
 
 
+(defparameter cl-user::*al-enable-breakpoints* t)
+(export 'cl-user::*al-enable-breakpoints* 'cl-user)
+
+(define-condition cl-user::al-break nil nil)
+(export 'cl-user::al-break 'cl-user)
+
+(defun break (&optional form &rest args)
+  (if cl-user:*al-enable-breakpoints*
+      (with-simple-restart (continue "")
+	(invoke-debugger (make-condition 'cl-user:al-break)))
+      nil))
+
+
+
+(defmacro ignore-errors (&rest forms)
+  `(handler-case (progn . ,forms)
+     (error (c) (values nil c))))
+
+
+
 (defun abort (&optional cond)
   (invoke-restart 'abort))
 
@@ -2743,6 +3045,16 @@
 
 
 
+(defgeneric documentation (x doc-type))
+
+(defmethod documentation ((x t) (doc-type t)))
+
+(defgeneric (setf documentation) (newval x doc-type))
+
+(defmethod (setf documentation) ((newval t) (x t) (doc-type t)))
+
+
+
 
 (export '(*features* *default-pathname-defaults* machine-instance machine-type
 	  machine-version short-site-name long-site-name *modules* provide
@@ -2750,8 +3062,8 @@
 	  call-arguments-limit multiple-values-limit array-rank-limit
 	  array-dimension-limit array-total-size-limit char-code-limit
 	  lambda-list-keywords identity constantly *read-default-float-format*
-	  pi 1+ 1- minusp plusp abs zerop signum mod rem evenp oddp gcd lcm
-	  isqrt conjugate cis upgraded-complex-part-type logand logandc1
+	  pi 1+ 1- minusp plusp abs zerop signum mod rem evenp oddp rationalize
+	  gcd lcm isqrt conjugate cis upgraded-complex-part-type logand logandc1
 	  logandc2 logeqv lognand lognor logorc1 logorc2 logxor boole-1 boole-2
 	  boole-andc1 boole-andc2 boole-and boole-c1 boole-c2 boole-clr
 	  boole-eqv boole-ior boole-nand boole-nor boole-orc1 boole-orc2
@@ -2760,22 +3072,23 @@
 	  cdar cddr caaar caadr cadar caddr cdaar cdadr cddar cdddr caaaar
 	  caaadr caadar caaddr cadaar cadadr caddar cadddr cdaaar cdaadr cdadar
 	  cdaddr cddaar cddadr cdddar cddddr make-list copy-alist copy-tree
-	  tree-equal sublis nsublis endp butlast nbutlast acons pairlis defsetf
-	  when unless define-modify-macro incf decf and or cond otherwise case
-	  ccase ecase typecase ctypecase etypecase return multiple-value-bind
-	  prog prog* every some notany notevery member member-if member-if-not
-	  find find-if find-if-not assoc assoc-if assoc-if-not rassoc rassoc-if
-	  rassoc-if-not position position-if position-if-not count count-if
-	  count-if-not remove remove-if-not delete delete-if delete-if-not
-	  remove-duplicates delete-duplicates substitute substitute-if
-	  substitute-if-not nsubstitute nsubstitute-if nsubstitute-if-not subst
-	  subst-if subst-if-not nsubst nsubst-if nsubst-if-not nreverse
-	  revappend nreconc adjoin fill replace push pushnew pop set-difference
-	  nset-difference union nunion intersection nintersection
-	  set-exclusive-or nset-exclusive-or subsetp mismatch search sort
-	  stable-sort array-rank array-dimension array-total-size
+	  tree-equal sublis nsublis endp butlast nbutlast acons pairlis shiftf
+	  rotatef defsetf when unless define-modify-macro incf decf cond
+	  otherwise case ccase ecase typecase ctypecase etypecase return
+	  multiple-value-bind multiple-value-setq prog prog*
+	  multiple-value-prog1 every some notany notevery member member-if
+	  member-if-not find find-if find-if-not assoc assoc-if assoc-if-not
+	  rassoc rassoc-if rassoc-if-not position position-if position-if-not
+	  count count-if count-if-not remove remove-if-not delete delete-if
+	  delete-if-not remove-duplicates delete-duplicates substitute
+	  substitute-if substitute-if-not nsubstitute nsubstitute-if
+	  nsubstitute-if-not subst subst-if subst-if-not nsubst nsubst-if
+	  nsubst-if-not nreverse revappend nreconc adjoin fill replace push
+	  pushnew pop set-difference nset-difference union nunion intersection
+	  nintersection set-exclusive-or nset-exclusive-or subsetp mismatch
+	  search sort stable-sort array-rank array-dimension array-total-size
 	  array-in-bounds-p array-element-type upgraded-array-element-type
-	  adjustable-array-p get get-properties remprop char schar bit sbit
+	  adjustable-array-p get get-properties remprop getf char schar bit sbit
 	  svref vector-pop vector-push vector-push-extend string= string/=
 	  string< string<= string> string>= string-equal string-not-equal
 	  string-lessp string-not-greaterp string-greaterp string-not-lessp
@@ -2790,14 +3103,17 @@
 	  stringp simple-string-p bit-vector-p simple-bit-vector-p hash-table-p
 	  pathnamep streamp realp numberp check-type assert macroexpand equal
 	  fdefinition complement mapc mapcan maplist mapl mapcon map-into reduce
-	  merge logical-pathname pathname-host pathname-device
-	  pathname-directory pathname-type pathname-version namestring
-	  file-namestring directory-namestring host-namestring enough-namestring
-	  merge-pathnames file-author file-write-date user-homedir-pathname
-	  with-open-file terpri write-line write-sequence prin1 princ print
-	  write-to-string prin1-to-string princ-to-string do-all-symbols loop
-	  format encode-universal-time with-standard-io-syntax handler-case
-	  restart-case with-simple-restart find-restart break abort continue))
+	  merge logical-pathname translate-logical-pathname pathname-host
+	  pathname-device pathname-version namestring file-namestring
+	  directory-namestring host-namestring enough-namestring merge-pathnames
+	  file-author file-write-date user-homedir-pathname with-open-file
+	  terpri write-line write-sequence prin1 princ print write-to-string
+	  prin1-to-string princ-to-string force-output with-input-from-string
+	  with-output-to-string pprint do-all-symbols find-all-symbols
+	  with-slots with-accessors loop format encode-universal-time
+	  with-compilation-unit *readtable* with-standard-io-syntax handler-case
+	  restart-case with-simple-restart find-restart cerror break
+	  ignore-errors abort continue documentation))
 
 
 
